/* Copy buttons */
button.copybtn {
  webkit-transition: opacity .3s ease-in-out;
  -o-transition: opacity .3s ease-in-out;
  transition: opacity .3s ease-in-out;
  opacity: 0;
  padding: 2px 6px;
  position: absolute;
  right: 4px;
  top: 4px;
}
div.highlight:hover .copybtn, div.highlight .copybtn:focus {
    opacity: .3;
}
div.highlight .copybtn:hover {
    opacity: 1;
}
div.highlight {
    position: relative;
}
